function [Zmap,p,q]=Stereo(image1,ps1,qs1,image2,ps2,qs2,image3,ps3,qs3,vps,vqs)

%Compute max of each irradiance image and take it as illumination intensity
%of sources

E1 = -255.0;
E2 = -255.0;
E3 = -255.0;

image1=double(image1);
image2=double(image2);
image3=double(image3);

[szy,szx]=size(image1);

for i=1:szy,
    for j=1:szx,
        if image1(i,j) > E1
            E1 = image1(i,j);
        end
        if image2(i,j) > E2
            E2 = image2(i,j);
        end
        if image2(i,j) > E3
            E3 = image3(i,j);
        end
    end
end

E=zeros(1,3);
ps=zeros(1,3);
qs=zeros(1,3);
ns=zeros(1,3);
E0=zeros(1,3);
sn1 = sqrt(ps1*ps1 + qs1*qs1 + 1.0);
sn2 = sqrt(ps2*ps2 + qs2*qs2 + 1.0);
sn3 = sqrt(ps3*ps3 + qs3*qs3 + 1.0);
ns(1)=sqrt(ps1*ps1 + qs1*qs1 + 1.0);
ns(2)=sqrt(ps2*ps2 + qs2*qs2 + 1.0);
ns(3)=sqrt(ps3*ps3 + qs3*qs3 + 1.0);

E0(1)=E1;
E0(2)=E2;
E0(3)=E3;

r = zeros(szy,szx);

ps(1)=ps1;
ps(2)=ps2;
ps(3)=ps3;

qs(1)=qs1;
qs(2)=qs2;
qs(3)=qs3;

thres=1; % threshold

p = zeros(szy,szx);
q = zeros(szy,szx);
A = zeros(szy,szx); %Calculate the Albedo map

pt=-0.12;
qt=0.34;
At=0.25;

for i=1:szy,
    for j=1:szx
        p(i,j) = 0.0;
        q(i,j) = 0.0;
        A(i,j) = 0.0;

        E(1) = image1(i,j);
        E(2) = image2(i,j);
        E(3) = image3(i,j);

        if image1(i,j) > thres | image2(i,j) > thres | image3(i,j) > thres
            %Lets try the three possible cross products for better estimates
            [pt,qt,At,ret]=computeNormal(E,E0,ps,qs,ns,1,2,3,thres,pt,qt,At);
            if ret==1
                p(i,j) = pt;
                q(i,j) = qt;
                A(i,j) = At;
            else
                [pt,qt,At,ret] = computeNormal(E,E0,ps,qs,ns,2,1,3,thres,pt,qt,At);
                if ret==1
                    p(i,j) = pt;
                    q(i,j) = qt;
                    A(i,j) = At;
                else
                    [pt,qt,At,ret] = computeNormal(E,E0,ps,qs,ns,3,2,1,thres,pt,qt,At);
                    if ret==1
                        p(i,j) = pt;
                        q(i,j) = qt;
                        A(i,j) = At;
                    end
                end
            end

            N = sqrt(p(i,j)*p(i,j) + q(i,j)*q(i,j) + 1.0);
            p(i,j) = -p(i,j)/N;
            q(i,j) = -q(i,j)/N;
            r(i,j) = 1.0/N;
        end
    end
end

% Integrability Check
diff=zeros(szy,szx);


for i=1:szy-1,
    for j=1:szx-1
        if i+1==szy | j+1 >=szx
            px=0.0;
            qx=0.0;
        else
            py = (p(i+1,j) - p(i,j))/2.0;
            qx = (q(i,j+1) - q(i,j))/2.0;
        end
        diff(i,j) = py - qx;
    end
end

imwrite(diff,'Integrability.pgm','pgm');

imwrite(A,'Albedo.pgm','pgm');

% Code for fake light source (or View Source)

N = sqrt(vps*vps + vqs*vqs +1.0);

fakeI=zeros(szy,szx);
for i=1:szy,
    for j=1:szx
        dp = (-p(i,j)*vps + q(i,j)*vqs + r(i,j))/N;
        fakeI(i,j)=dp*1.0*A(i,j);

    end
end


imwrite(fakeI,'fakeI.pgm','pgm');

[Zmap]=ShapeFromGradient(p,q);



function [pt,qt,At,ret]=computeNormal(E,E0,ps,qs,ns,i1,i2,i3,T,pt,qt,At)

if (E(i1) <=T)
    ret=0;
else
    
    alpha12 = (ns(i2)/ns(i1)) * (E(i2)/E(i1)) * (E0(i1)/E0(i2));
    alpha13 = (ns(i3)/ns(i1)) * (E(i3)/E(i1)) * (E0(i1)/E0(i3));
    %alpha12=(ns(i2)/ns(i1))*(E(i2))*(E0(i1)/E0(i2));
    %alpha13=(ns(i3)/ns(i1))*(E(i3))*(E0(i1)/E0(i3));
    
    a = ps(i2) - alpha12*ps(i1);
    b = qs(i2) - alpha12*qs(i1);
    c = 1.0 - alpha12;
    d = ps(i3) - alpha13*ps(i1);
    e = qs(i3) - alpha13*qs(i1);
    f = 1.0 - alpha13;
    
%    wait;
    N = a*e - b*d; %z componenet of surface normal (non zero)
    if N==0
        disp('Warning: Unconsistent normal component');
        ret=0;
    else
        % Computation of p and q (cross product of S12 and S13)
        pt = (b*f - e*c)/N;
        qt = (d*c - a*f)/N;

        % Computation of the albedo
        dp = ((pt*ps(i1) + qt*qs(i1) + 1.0)*E0(i1))/(ns(i1)*(sqrt(pt*pt + qt*qt + 1.0)));
        At  = E(i1)/dp;
        if At > 1
            At = 1.0;
        end
        if At < 0
            At = 0.0;
        end
        ret=1;
    end
end


function [Zmap]=ShapeFromGradient(p,q)
[szy,szx]=size(p);
z1=zeros(szy,szx);
z2=zeros(szy,szx);
z3=zeros(szy,szx);
z4=zeros(szy,szx);


% First scan (top left) a|b
%                       ---
%                       c|d
for i=2:szy
    for j=2:szx
        nx = (p(i-1,j-1) + p(i-1,j) + p(i,j-1) + p(i,j))/4.0;
        ny = (q(i-1,j-1) + q(i-1,j) + q(i,j-1) + q(i,j))/4.0;
        nz = -sqrt(1.0 - nx*nx - ny*ny);
        dx = -nx/nz; 
        dy = ny/nz;
        z1(i,j) = (z1(i-1,j) + z1(i,j-1))/2.0 + (dx + dy)/2.0;
    end
end

% Second scan (bottom right) d|c
%                            ---
%                            b|a

for i=szy-1:-1:1
    for j=szx-1:-1:1
        nx = (p(i+1,j+1) + p(i+1,j) + p(i,j+1) + p(i,j))/4.0;
        ny = (q(i+1,j+1) + q(i+1,j) + q(i,j+1) + q(i,j))/4.0;
        nz = -sqrt(1.0 - nx*nx - ny*ny);
        dx = nx/nz;
        dy = -ny/nz;
        z2(i,j) = (z2(i+1,j) + z2(i,j+1))/2.0 + (dx + dy)/2.0;
    end
end


% Third scan (top right)  c|a
%                         ---
%                         d|b

for j=szx-1:-1:2
    for i=2:szy
        nx = (p(i-1,j+1) + p(i,j+1) + p(i-1,j) + p(i,j))/4.0;
        ny = (q(i-1,j+1) + q(i,j+1) + q(i-1,j) + q(i,j))/4.0;
        nz = -sqrt(1.0 - nx*nx - ny*ny);
        dx = nx/nz; 
        dy = ny/nz;
        z3(i,j) = (z3(i,j+1) + z3(i-1,j))/2.0 + (dx + dy)/2.0;
    end
end

% Fourth scan (bottom left)  b|d
%                            ---
%                            a|c

for j=2:szx
    for i=szy-1:1,-1
        nx = (p(i+1,j-1) + p(i,j-1) + p(i+1,j) + p(i,j))/4.0;
        ny = (q(i+1,j-1) + q(i,j-1) + q(i+1,j) + q(i,j))/4.0;
        nz = -sqrt(1.0 - nx*nx - ny*ny);
        dx = -nx/nz;
        dy = -ny/nz;
        
        z4(i,j) = (z4(i,j-1) + z4(i+1,j))/2.0 + (dx + dy)/2.0;
    end
end

Zmap=zeros(szy,szx);

for i=1:szy,
    for j=1:szx
        %Zmap(i,j) = sqrt((z1(i,j)*z1(i,j) + z2(i,j)*z2(i,j) + z3(i,j)*z3(i,j) + z4(i,j)*z4(i,j))/4.0);
        Zmap(i,j) = (z1(i,j) + z2(i,j) + z3(i,j)+ z4(i,j))/4;
    end
end



